﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using GMap.NET;
using Schedulr.Models;

namespace Schedulr.Views.Controls
{
    /// <summary>
    /// Interaction logic for GeoLocationMap.xaml
    /// </summary>
    public partial class GeoLocationMap : UserControl
    {
        #region Fields

        private static IEnumerable<MapTypeViewModel> availableMaps;

        #endregion

        #region Constructors

        static GeoLocationMap()
        {
            availableMaps = new MapTypeViewModel[]
            {
                new MapTypeViewModel(GeoLocationMapType.BingRoad, MapType.BingMap_New, "Bing Map (Road)"),
                new MapTypeViewModel(GeoLocationMapType.BingSattelite, MapType.BingSatellite, "Bing Map (Satellite)"),
                new MapTypeViewModel(GeoLocationMapType.BingHybrid, MapType.BingHybrid, "Bing Map (Hybrid)"),
                new MapTypeViewModel(GeoLocationMapType.YahooRoad, MapType.YahooMap, "Yahoo Map (Road)"),
                new MapTypeViewModel(GeoLocationMapType.YahooSatellite, MapType.YahooSatellite, "Yahoo Map (Satellite)"),
                new MapTypeViewModel(GeoLocationMapType.YahooHybrid, MapType.YahooHybrid, "Yahoo Map (Hybrid)"),
                new MapTypeViewModel(GeoLocationMapType.OpenStreetMapRoad, MapType.OpenStreetMap, "OpenStreetMap (Road)")
            };
        }

        public GeoLocationMap()
        {
            InitializeComponent();

            var defaultMapType = availableMaps.First();
            this.geoGMapControl.Manager.Mode = AccessMode.ServerOnly;
            this.geoGMapControl.MapType = defaultMapType.GMapMapType;
            this.geoGMapControl.DragButton = MouseButton.Left;
            this.geoGMapControl.MouseWheelZoomType = MouseWheelZoomType.MousePositionWithoutCenter;
            this.geoGMapControl.MinZoom = 1;
            this.geoGMapControl.MaxZoom = GeoLocation.MaxAccuracy;
            this.geoGMapControl.Zoom = 1;
            this.geoGMapControl.OnMapZoomChanged += new MapZoomChanged(geoGMapControl_OnMapZoomChanged);
            this.geoGMapControl.OnCurrentPositionChanged += new CurrentPositionChanged(geoGMapControl_OnCurrentPositionChanged);
            this.geoGMapControl.Manager.UseGeocoderCache = false;

            this.mapTypesComboBox.ItemsSource = availableMaps;
            this.mapTypesComboBox.SelectedItem = defaultMapType;
            this.mapTypesComboBox.SelectionChanged += new SelectionChangedEventHandler(mapTypesComboBox_SelectionChanged);
        }

        #endregion

        #region Event Handlers

        private void geoGMapControl_OnCurrentPositionChanged(PointLatLng point)
        {
            this.Location.Latitude = point.Lat;
            this.Location.Longitude = point.Lng;
        }

        private void geoGMapControl_OnMapZoomChanged()
        {
            // Even though the map is configured with min and max zoom values, it can still return values outside that range.
            // Explicitly constrain the map's zoom level to the allowed values for a GeoLocation.
            var zoomLevel = (int)this.geoGMapControl.Zoom;
            zoomLevel = Math.Max(GeoLocation.MinAccuracy, Math.Min(GeoLocation.MaxAccuracy, zoomLevel));
            this.Location.Accuracy = zoomLevel;
        }

        private void mapTypesComboBox_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            this.GeoLocationMapType = ((MapTypeViewModel)this.mapTypesComboBox.SelectedItem).MapType;
        }

        #endregion

        #region IsReadOnly

        public bool IsReadOnly
        {
            get { return (bool)GetValue(IsReadOnlyProperty); }
            set { SetValue(IsReadOnlyProperty, value); }
        }

        public static readonly DependencyProperty IsReadOnlyProperty = DependencyProperty.Register("IsReadOnly", typeof(bool), typeof(GeoLocationMap), new UIPropertyMetadata(false, OnIsReadOnlyChanged));

        private static void OnIsReadOnlyChanged(DependencyObject o, DependencyPropertyChangedEventArgs e)
        {
            var instance = (GeoLocationMap)o;
            instance.geoGMapControl.IsEnabled = !instance.IsReadOnly;
            instance.zoomStackPanel.Visibility = (instance.IsReadOnly ? Visibility.Collapsed : Visibility.Visible);
            instance.mapTypesComboBox.Visibility = (instance.IsReadOnly ? Visibility.Collapsed : Visibility.Visible);
        }

        #endregion

        #region Location

        public GeoLocation Location
        {
            get { return (GeoLocation)GetValue(LocationProperty); }
            set { SetValue(LocationProperty, value); }
        }

        public static readonly DependencyProperty LocationProperty = DependencyProperty.Register("Location", typeof(GeoLocation), typeof(GeoLocationMap), new UIPropertyMetadata(new GeoLocation(), OnLocationChanged));

        private static void OnLocationChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
        {
            var instance = (GeoLocationMap)sender;

            var oldLocation = e.OldValue as GeoLocation;
            if (oldLocation != null)
            {
                oldLocation.PropertyChanged -= instance.Location_PropertyChanged;
            }

            if (instance.Location == null)
            {
                instance.Location = new GeoLocation();
            }
            instance.Location.PropertyChanged += instance.Location_PropertyChanged;
            instance.UpdateMap();
        }

        private void Location_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            UpdateMap();
        }

        private void UpdateMap()
        {
            this.geoGMapControl.Position = new PointLatLng(this.Location.Latitude, this.Location.Longitude);
            this.geoGMapControl.Zoom = Math.Max(1, (double)this.Location.Accuracy);
        }

        #endregion

        #region SearchLocation

        /// <summary>
        /// Searches for a location and sets the current position if found.
        /// </summary>
        /// <param name="searchText">The location to search for.</param>
        /// <returns><see langword="true"/> if the location was found, <see langword="false"/> otherwise.</returns>
        public bool SearchLocation(string searchText)
        {
            var status = this.geoGMapControl.SetCurrentPositionByKeywords(searchText);
            return (status == GeoCoderStatusCode.G_GEO_SUCCESS);
        }

        #endregion

        #region GeoLocationMapType

        public GeoLocationMapType GeoLocationMapType
        {
            get { return (GeoLocationMapType)GetValue(GeoLocationMapTypeProperty); }
            set { SetValue(GeoLocationMapTypeProperty, value); }
        }

        public static readonly DependencyProperty GeoLocationMapTypeProperty = DependencyProperty.Register("GeoLocationMapType", typeof(GeoLocationMapType), typeof(GeoLocationMap), new UIPropertyMetadata(OnGeoLocationMapTypeChanged));

        private static void OnGeoLocationMapTypeChanged(DependencyObject o, DependencyPropertyChangedEventArgs e)
        {
            var instance = (GeoLocationMap)o;
            var mapTypeViewModel = availableMaps.Where(m => m.MapType == (GeoLocationMapType)e.NewValue).FirstOrDefault() ?? availableMaps.First();
            instance.mapTypesComboBox.SelectedItem = mapTypeViewModel;
            instance.geoGMapControl.MapType = mapTypeViewModel.GMapMapType;
        }

        #endregion

        #region Private MapTypeViewModel Class

        private class MapTypeViewModel
        {
            public GeoLocationMapType MapType { get; private set; }
            public MapType GMapMapType { get; private set; }
            public string Name { get; private set; }

            public MapTypeViewModel(GeoLocationMapType mapType, MapType gMapMapType, string name)
            {
                this.MapType = mapType;
                this.GMapMapType = gMapMapType;
                this.Name = name;
            }
        }

        #endregion
    }
}