using System;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Text;
using System.Xml;
using Schedulr.Configuration;
using Schedulr.Properties;

namespace Schedulr.Helpers
{
    /// <summary>
    /// Provides helper methods for pictures.
    /// </summary>
    internal static class PictureHelper
    {
        #region GetPicture

        /// <summary>
        /// Gets the picture information for a picture file.
        /// </summary>
        /// <param name="fullPath">The full path to the picture file.</param>
        /// <returns>The picture information for the given path.</returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        public static Picture GetPicture(string fullPath)
        {
            Picture pictureToAdd = new Picture();
            pictureToAdd.FileName = fullPath;
            pictureToAdd.Title = Path.GetFileNameWithoutExtension(fullPath);

            if (Settings.Default.RetrieveMetaData)
            {
                // Try to get the metadata we need.
                try
                {
                    Logger.Log("Retrieving metadata from picture " + fullPath, TraceEventType.Information);
                    string fileContents = null;
                    using (StreamReader reader = new StreamReader(fullPath))
                    {
                        fileContents = reader.ReadToEnd();
                    }
                    const string beginCapture = "<rdf:RDF";
                    const string endCapture = "</rdf:RDF>";
                    int beginPos = fileContents.IndexOf(beginCapture, StringComparison.OrdinalIgnoreCase);
                    int endPos = fileContents.IndexOf(endCapture, StringComparison.OrdinalIgnoreCase);
                    if (beginPos >= 0 && endPos > beginPos)
                    {
                        string xmlPart = fileContents.Substring(beginPos, (endPos - beginPos) + endCapture.Length);
                        RetrieveMetaData(xmlPart, pictureToAdd);
                    }
                    else
                    {
                        Logger.Log("Could not retrieve metadata from picture");
                    }
                }
                catch(Exception exc)
                {
                    Logger.Log("Could not retrieve metadata from picture", exc);
                }
            }
            return pictureToAdd;
        }

        #endregion

        #region Helper Methods

        /// <summary>
        /// Retrieves the meta data.
        /// </summary>
        /// <param name="embeddedMetaDataXml">The embedded meta data XML.</param>
        /// <param name="picture">The picture.</param>
        private static void RetrieveMetaData(string embeddedMetaDataXml, Picture picture)
        {
            // Thank you Omar: http://www.shahine.com/omar/ReadingXMPMetadataFromAJPEGUsingC.aspx.
            XmlDocument doc = new XmlDocument();
            doc.LoadXml(embeddedMetaDataXml);

            XmlNamespaceManager nsMgr = new XmlNamespaceManager(doc.NameTable);
            nsMgr.AddNamespace("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
            nsMgr.AddNamespace("exif", "http://ns.adobe.com/exif/1.0/");
            nsMgr.AddNamespace("x", "adobe:ns:meta/");
            nsMgr.AddNamespace("xap", "http://ns.adobe.com/xap/1.0/");
            nsMgr.AddNamespace("tiff", "http://ns.adobe.com/tiff/1.0/");
            nsMgr.AddNamespace("dc", "http://purl.org/dc/elements/1.1/");

            // get keywords
            XmlNode xmlNode = doc.SelectSingleNode("/rdf:RDF/rdf:Description/dc:subject/rdf:Bag", nsMgr);
            if (xmlNode != null)
            {
                StringBuilder tags = new StringBuilder();
                foreach (XmlNode tagNode in xmlNode)
                {
                    if (tags.Length > 0)
                    {
                        // Separate tags with spaces.
                        tags.Append(" ");
                    }
                    string tag = tagNode.InnerText;
                    if (tag.Contains(" "))
                    {
                        // If the tag contains spaces, put double quotes around them.
                        tag = string.Format(CultureInfo.CurrentCulture, "\"{0}\"", tag);
                    }
                    tags.Append(tag);
                }
                picture.Tags = tags.ToString();
            }

            // get title
            xmlNode = doc.SelectSingleNode("/rdf:RDF/rdf:Description/dc:title/rdf:Alt", nsMgr);
            if (xmlNode != null && xmlNode.ChildNodes.Count > 0)
            {
                picture.Title = xmlNode.ChildNodes[0].InnerText;
            }

            // get description
            xmlNode = doc.SelectSingleNode("/rdf:RDF/rdf:Description/dc:description/rdf:Alt", nsMgr);
            if (xmlNode != null && xmlNode.ChildNodes.Count > 0)
            {
                picture.Description = xmlNode.ChildNodes[0].InnerText;
            }
        }

        #endregion
    }
}