﻿using System;
using System.Collections.ObjectModel;
using System.IO;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Schedulr.Models;
using Schedulr.Providers;

namespace Schedulr.Test
{
    [TestClass]
    public class SchedulrConfigurationProviderTest
    {
        private const string TestConfigurationFile = "TestSchedulrConfiguration.xml";
        private const string TestConfigurationFileFromV1 = "Configuration-v1.4.xml";
        private const string TestConfigurationFileFromV2 = "Configuration-v2.0.xml";

        [TestMethod]
        public void SaveWritesAndLoadsConfigurationFile()
        {
            var writeConfig = GetTestConfiguration();

            // Save the configuration.
            Assert.IsFalse(File.Exists(TestConfigurationFile));
            SchedulrConfigurationProvider.Save(writeConfig);
            Assert.IsTrue(File.Exists(TestConfigurationFile));

            // Read the saved configuration and compare.
            var readConfig = SchedulrConfigurationProvider.Load();
            Verifier.VerifySchedulrConfiguration(writeConfig, readConfig);

            // Clean up.
            File.Delete(TestConfigurationFile);
        }

        [TestMethod]
        public void LoadOrCreateLoadsOrCreatesConfiguration()
        {
            // If the configuration file does not exist, creates a new one.
            Assert.IsFalse(File.Exists(TestConfigurationFile));
            var createdConfig = SchedulrConfigurationProvider.LoadOrCreate();
            Assert.IsNotNull(createdConfig);
            Assert.AreEqual<int>(0, createdConfig.Accounts.Count);

            // Save the test configuration to force a file with multiple accounts to be present.
            var testConfig = GetTestConfiguration();
            SchedulrConfigurationProvider.Save(testConfig);

            // Now the configuration file exists so it should be loaded.
            Assert.IsTrue(File.Exists(TestConfigurationFile));
            var loadedConfig = SchedulrConfigurationProvider.LoadOrCreate();
            Assert.IsNotNull(loadedConfig);
            Assert.AreEqual<int>(testConfig.Accounts.Count, loadedConfig.Accounts.Count);

            // Now make the file unreadable as a configuration file.
            File.WriteAllText(TestConfigurationFile, "Dummy");
            var fallbackConfig = SchedulrConfigurationProvider.LoadOrCreate();
            Assert.IsNotNull(fallbackConfig);
            Assert.AreEqual<int>(0, fallbackConfig.Accounts.Count);

            // Clean up.
            File.Delete(TestConfigurationFile);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void LoadThrowsOnNullFileName()
        {
            SchedulrConfigurationProvider.Load(null);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void LoadThrowsOnEmptyFileName()
        {
            SchedulrConfigurationProvider.Load(string.Empty);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void LoadThrowsOnNonExistentFileName()
        {
            SchedulrConfigurationProvider.Load(Util.GetDummyFileName());
        }

        [TestMethod]
        [DeploymentItem(TestConfigurationFileFromV1)]
        public void LoadImportsConfigurationFromV1()
        {
            var account = new Account();
            var count = SchedulrConfigurationProvider.ImportFromLegacyConfiguration(account, TestConfigurationFileFromV1);
            var equivalentPictures = GetEquivalentPicturesFromV1();
            Assert.AreEqual<int>(equivalentPictures.Count, count);
            Assert.AreEqual<int>(0, account.UploadedPictures.Count); // Uploaded pictures should NOT be imported.
            Verifier.VerifyPictures(equivalentPictures, account.QueuedPictures);
        }

        [TestMethod]
        [DeploymentItem(TestConfigurationFileFromV2)]
        public void MyTestMethod()
        {
            // Verify static config file with in-memory config (making sure that file format stays compatible over time).
            var testConfig = GetTestConfiguration();
            var readConfig = SchedulrConfigurationProvider.Load(TestConfigurationFileFromV2);
            Verifier.VerifySchedulrConfiguration(testConfig, readConfig);
        }

        private static SchedulrConfiguration GetTestConfiguration()
        {
            return new SchedulrConfiguration
            {
                Accounts =
                {
                    new Account
                    {
                        Id = "John-ID",
                        Name = "John",
                        IsDefaultAccount = true,
                        AuthenticationToken = "token-john",
                        QueuedPictures = new ObservableCollection<Picture>
                        {
                            new Picture
                            {
                                Title = "title-1",
                                Description = "description-1. This description spans\r\nmultiple lines\r\n\r\nwhich should be supported!",
                                Tags = "tags-1",
                                VisibilityIsFamily = true,
                                VisibilityIsFriend = true,
                                VisibilityIsPublic = true,
                                Safety = Safety.Safe,
                                ContentType = ContentType.Photo,
                                License = License.AttributionCC,
                                GroupIds = { "group-id-2", "group-id-3" },
                                SetIds = { "set-id-1" }
                            },
                            new Picture
                            {
                                Title = "title-2",
                                Description = "description-2. This description spans\r\nmultiple lines\r\n\r\nwhich should be supported!",
                                Tags = "tags-2",
                                VisibilityIsFamily = false,
                                VisibilityIsFriend = false,
                                VisibilityIsPublic = false,
                                Safety = Safety.Restricted,
                                ContentType = ContentType.None,
                                License = License.AttributionNoDerivativesCC,
                                SetIds = { "set-id-1", "set-id-2" }
                            }
                        },
                        UploadedPictures = new ObservableCollection<Picture>
                        {
                            new Picture
                            {
                                Title = "title-3",
                                Description = "description-3. This description spans\r\nmultiple lines\r\n\r\nwhich should be supported!",
                                Tags = "tags-3",
                                VisibilityIsFamily = true,
                                VisibilityIsFriend = false,
                                VisibilityIsPublic = true,
                                Safety = Safety.None,
                                ContentType = ContentType.Screenshot,
                                License = License.AttributionNoncommercialCC,
                                GroupIds = { "group-id-1", "group-id-3" },
                                SetIds = { "set-id-2" },
                                DateUploaded = new DateTimeOffset(2010, 06, 28, 20, 51, 23, TimeSpan.FromHours(2)).DateTime
                            }
                        },
                        UserInfo = new UserInfo
                        {
                            UserName = "John",
                            PicturesUrl = "http://www.flickr.com/photos/john",
                            BuddyIconUrl = "http://www.flickr.com/photos/john/buddy-icon",
                            IsProUser = true,
                            UploadLimit = 10 * 1024 * 1024,
                            Groups = 
                            {
                                new PictureCollection { Id = "group-id-1", Name = "group-name-1", ImageUrl = "group-image-1" },
                                new PictureCollection { Id = "group-id-2", Name = "group-name-2", ImageUrl = "group-image-2" },
                                new PictureCollection { Id = "group-id-3", Name = "group-name-3", ImageUrl = "group-image-3" }
                            },
                            Sets = 
                            {
                                new PictureCollection { Id = "set-id-1", Name = "set-name-1", ImageUrl = "set-image-1" },
                                new PictureCollection { Id = "set-id-2", Name = "set-name-2", ImageUrl = "set-image-2" }
                            }
                        },
                        Settings = new AccountSettings
                        {
                            DeleteFileAfterUpload = false,
                            RetrieveMetadata = true,
                            PictureDefaults = new Picture
                            {
                                Title = "default-title",
                                Description = "default-description. This description spans\r\nmultiple lines\r\n\r\nwhich should be supported!",
                                Tags = "default-tags",
                                VisibilityIsFamily = true,
                                VisibilityIsFriend = false,
                                VisibilityIsPublic = true,
                                Safety = Safety.Safe,
                                ContentType = ContentType.Photo,
                                License = License.AttributionNoncommercialNoDerivativesCC,
                                GroupIds = { "group-id-2", "group-id-3" },
                                SetIds = { "set-id-1" }
                            }
                        },
                        UploadSchedule = new UploadSchedule
                        {
                            Type = ScheduleType.Daily,
                            StartTime = new DateTimeOffset(2010, 06, 30, 19, 00, 00, TimeSpan.FromHours(2)).DateTime,
                            OnlyIfLoggedOn = false,
                            WakeComputer = true,
                            OnlyIfNotOnBatteryPower = true,
                            HourlyInterval = 2,
                            HourlyIntervalDuration = 4,
                            DailyInterval = 3,
                            WeeklyInterval = 4,
                            WeeklyOnMonday = false,
                            WeeklyOnTuesday = true,
                            WeeklyOnWednesday = false,
                            WeeklyOnThursday = true,
                            WeeklyOnFriday = false,
                            WeeklyOnSaturday = true,
                            WeeklyOnSunday = false
                        }
                    },
                    new Account
                    {
                        Id = "Mary-ID",
                        Name = "Mary",
                        IsDefaultAccount = false,
                        AuthenticationToken = "token-mary",
                        UserInfo = new UserInfo
                        {
                            UserName = "Mary",
                            PicturesUrl = "http://www.flickr.com/photos/mary",
                            BuddyIconUrl = "http://www.flickr.com/photos/mary/buddy-icon",
                            IsProUser = false,
                            UploadLimit = long.MaxValue,
                            Groups = 
                            {
                                new PictureCollection { Id = "group-id-2", Name = "group-name-2", ImageUrl = "group-image-2" },
                                new PictureCollection { Id = "group-id-4", Name = "group-name-4", ImageUrl = "group-image-4" }
                            },
                        },
                        Settings = new AccountSettings
                        {
                            DeleteFileAfterUpload = true,
                            RetrieveMetadata = false,
                            PictureDefaults = new Picture
                            {
                                Title = "default-title-2",
                                Description = "default-description-2. This description spans\r\nmultiple lines\r\n\r\nwhich should be supported!",
                                Tags = "default-tags-2",
                                VisibilityIsFamily = false,
                                VisibilityIsFriend = true,
                                VisibilityIsPublic = false,
                                Safety = Safety.Restricted,
                                ContentType = ContentType.Other,
                                License = License.AllRightsReserved,
                                GroupIds = { "group-id-4" }
                            }
                        },
                        UploadSchedule = new UploadSchedule
                        {
                            Type = ScheduleType.Weekly,
                            StartTime = new DateTimeOffset(2010, 07, 01, 03, 00, 00, TimeSpan.FromHours(2)).DateTime,
                            OnlyIfLoggedOn = true,
                            WakeComputer = false,
                            OnlyIfNotOnBatteryPower = false,
                            HourlyInterval = 6,
                            HourlyIntervalDuration = 8,
                            DailyInterval = 7,
                            WeeklyInterval = 8,
                            WeeklyOnMonday = true,
                            WeeklyOnTuesday = false,
                            WeeklyOnWednesday = true,
                            WeeklyOnThursday = false,
                            WeeklyOnFriday = true,
                            WeeklyOnSaturday = false,
                            WeeklyOnSunday = true
                        }
                    }
                }
            };
        }

        private static ObservableCollection<Picture> GetEquivalentPicturesFromV1()
        {
            // Create the equivalent of the pictures in the legacy v1 config file.
            return new ObservableCollection<Picture>
            {
                new Picture
                {
                    FileName = @"C:\Users\Schedulr\Pictures\Upload\First Day.jpg",
                    Tags = "\"Day 1\" Travel Train",
                    VisibilityIsPublic = true,
                    VisibilityIsFamily = false,
                    VisibilityIsFriend = false,
                    SetIds = { "set-1" },
                    GroupIds = { "group-1" },
                    Title = "Arrival",
                    Description = @"This was a great first day.

Really!",
                    BatchId = 1
                },
                new Picture
                {
                    FileName = @"C:\Users\Schedulr\Pictures\Upload\Second Day 1.jpg",
                    Tags = "\"Day 2\" Travel Pool",
                    VisibilityIsPublic = false,
                    VisibilityIsFamily = true,
                    VisibilityIsFriend = false,
                    SetIds = { "set-2" },
                    GroupIds = { "group-2" },
                    Title = "Swimming Pool",
                    Description = "Chilling in the pool!",
                    BatchId = 2
                },
                new Picture
                {
                    FileName = @"C:\Users\Schedulr\Pictures\Upload\Second Day 2.jpg",
                    Tags = "\"Day 2\" Travel Beach",
                    VisibilityIsPublic = false,
                    VisibilityIsFamily = true,
                    VisibilityIsFriend = false,
                    SetIds = { "set-2" },
                    GroupIds = { "group-2" },
                    Title = "Beach",
                    Description = "Sunbathing on the beach...",
                    BatchId = 2
                },
                new Picture
                {
                    FileName = @"C:\Users\Schedulr\Pictures\Upload\Third Day 1.jpg",
                    Tags = "\"Day 3\" Travel City",
                    VisibilityIsPublic = false,
                    VisibilityIsFamily = false,
                    VisibilityIsFriend = true,
                    SetIds = { "set-3" },
                    GroupIds = { "group-3" },
                    Title = "City Tripping",
                    Description = "Checking out the city.",
                    BatchId = 3
                },
                new Picture
                {
                    FileName = @"C:\Users\Schedulr\Pictures\Upload\Third Day 2.jpg",
                    Tags = "\"Day 3\" Travel Museum",
                    VisibilityIsPublic = false,
                    VisibilityIsFamily = false,
                    VisibilityIsFriend = true,
                    SetIds = { "set-3" },
                    GroupIds = { "group-3" },
                    Title = "Museum",
                    Description = "A bit of culture.",
                    BatchId = 3
                },
                new Picture
                {
                    FileName = @"C:\Users\Schedulr\Pictures\Upload\Fourth Day.jpg",
                    Tags = "\"Day 4\" Travel Plane",
                    VisibilityIsPublic = true,
                    VisibilityIsFamily = true,
                    VisibilityIsFriend = true,
                    SetIds = { "set-1", "set-2", "set-3", "set-4" },
                    GroupIds = { "group-1", "group-2", "group-3", "group-4" },
                    Title = "Departure",
                    Description = string.Empty,
                    BatchId = 4
                },
            };
        }
    }
}