﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Windows;
using System.Windows.Input;
using System.Windows.Markup;
using System.Windows.Shell;
using JelleDruyts.Windows;
using Microsoft.Win32;
using Schedulr.Infrastructure;
using Schedulr.Messages;
using Schedulr.Models;
using Schedulr.Providers;
using Schedulr.Views;

namespace Schedulr
{
    /// <summary>
    /// Interaction logic for App.xaml
    /// </summary>
    public partial class App : Application
    {
        #region Constants & Static Properties

        /// <summary>
        /// Gets the URL where the application lives.
        /// </summary>
        public const string ApplicationUrl = "http://schedulr.codeplex.com";

        /// <summary>
        /// Gets the name of the application.
        /// </summary>
        public const string ApplicationName = Constants.ApplicationName;

        /// <summary>
        /// Gets the default title for the main window.
        /// </summary>
        public const string MainWindowTitle = ApplicationName;

        /// <summary>
        /// Gets the display version of the application.
        /// </summary>
        public const string DisplayVersion = "v2.3";

        /// <summary>
        /// Gets the full technical version of the application.
        /// </summary>
        public static Version FullVersion { get { return Assembly.GetExecutingAssembly().GetName().Version; } }

        /// <summary>
        /// Gets the executable name of the application.
        /// </summary>
        public static string ExecutableName { get { return Assembly.GetExecutingAssembly().GetName().Name; } }

        /// <summary>
        /// Gets the path to the executable of the application.
        /// </summary>
        public static string ExecutablePath { get { return Assembly.GetExecutingAssembly().Location; } }

        /// <summary>
        /// Gets the singleton <see cref="App"/> instance.
        /// </summary>
        public static App Instance { get { return (App)Application.Current; } }

        /// <summary>
        /// The known file extensions for video files.
        /// </summary>
        private static readonly IList<string> VideoFileExtensions = new string[] { ".avi", ".wmv", ".mov", ".mpg", ".mpeg", ".3gp", ".m2ts", ".ogg", ".ogv" };

        #endregion

        #region Properties

        public RelayCommand OpenLogFileCommand { get; private set; }
        public RelayCommand SaveChangesCommand { get; private set; }
        public RelayCommand UndoChangesCommand { get; private set; }
        public RelayCommand ExportConfigurationCommand { get; private set; }
        public RelayCommand ImportConfigurationCommand { get; private set; }

        private SchedulrConfiguration configuration;
        public SchedulrConfiguration Configuration
        {
            get
            {
                return this.configuration;
            }
            private set
            {
                if (value != this.configuration)
                {
                    this.configuration = value;
                    Messenger.Send<ConfigurationChangedMessage>(new ConfigurationChangedMessage(this.configuration));
                }
            }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="App"/> class.
        /// </summary>
        /// <exception cref="T:System.InvalidOperationException">More than one instance of the <see cref="T:System.Windows.Application"/> class is created per <see cref="T:System.AppDomain"/>.</exception>
        public App()
        {
            this.OpenLogFileCommand = new RelayCommand(OpenLogFile, CanOpenLogFile, "Open _Log File", "Opens the file that contains the activity log for this application [CTRL-L]", new KeyGesture(Key.L, ModifierKeys.Control));
            this.SaveChangesCommand = new RelayCommand(SaveChanges, CanSaveChanges, "_Save All Changes", "Saves the current configuration. This is done automatically when closing the application, but it can be safer to save once in a while or before trying some things out [CTRL-S]", new KeyGesture(Key.S, ModifierKeys.Control));
            this.UndoChangesCommand = new RelayCommand(UndoChanges, CanUndoChanges, "_Undo All Changes", "Reloads the last saved configuration. Since the application automatically saves all changes when it closes, this allows you to undo changes before closing [CTRL-Z]", new KeyGesture(Key.Z, ModifierKeys.Control));
            this.ExportConfigurationCommand = new RelayCommand(ExportConfiguration, CanExportConfiguration, "E_xport Configuration...", "Exports the current configuration to a file [CTRL-X]", new KeyGesture(Key.X, ModifierKeys.Control));
            this.ImportConfigurationCommand = new RelayCommand(ImportConfiguration, CanImportConfiguration, "_Import Configuration...", "Imports a configuration file (this will overwrite all your changes) [CTRL-I]", new KeyGesture(Key.I, ModifierKeys.Control));
            this.InitializeComponent();
        }

        #endregion

        #region Initialize

        /// <summary>
        /// Initializes the application.
        /// </summary>
        /// <param name="configuration">The configuration.</param>
        /// <param name="selectedAccount">The selected account from the command line arguments.</param>
        /// <param name="uploadUIRequested">Determines if the upload confirmation dialog was requested.</param>
        public void Initialize(SchedulrConfiguration configuration, Account selectedAccount, bool uploadUIRequested)
        {
            // Ensure that the current culture is used for all controls (see http://www.west-wind.com/Weblog/posts/796725.aspx).
            FrameworkElement.LanguageProperty.OverrideMetadata(typeof(FrameworkElement), new FrameworkPropertyMetadata(XmlLanguage.GetLanguage(CultureInfo.CurrentCulture.IetfLanguageTag)));

            // Create and show the main window.
            var main = new MainWindow();
            main.ContentRendered += delegate
            {
                var loadingTask = new ApplicationTask("Loading configuration");
                try
                {
                    // Set configuration.
                    Messenger.Send<StatusMessage>(new TaskStatusMessage(loadingTask));
                    this.Configuration = configuration;

                    // Set up jump list.
                    SetJumpList();

                    // Show the upload confirmation dialog if requested.
                    if (uploadUIRequested && selectedAccount != null)
                    {
                        Messenger.Send<UploadPicturesRequestedMessage>(new UploadPicturesRequestedMessage(selectedAccount.QueuedPictures.GetLowestBatch(), selectedAccount, UploadPicturesRequestReason.CommandLineUI));
                    }
                }
                finally
                {
                    loadingTask.SetComplete(this.Configuration.Accounts.Count.ToCountString("account", null, " loaded"));
                }
            };
            main.Closing += delegate(object sender, CancelEventArgs args)
            {
                // Send closing message to allow cancellation.
                var message = new ApplicationClosingMessage();
                Messenger.Send<ApplicationClosingMessage>(message);
                args.Cancel = message.Cancel;
            };
            main.Closed += delegate
            {
                // Update jump list (if accounts were changed).
                SetJumpList();

                // Save changes.
                SaveChanges(null);

                // Close.
                Messenger.Send<ApplicationClosedMessage>(new ApplicationClosedMessage());
            };
            Messenger.Send<ApplicationLoadedMessage>(new ApplicationLoadedMessage());
            main.Show();
        }

        private void SetJumpList()
        {
            var jumpList = new JumpList();
            foreach (var account in this.Configuration.Accounts)
            {
                var task = new JumpTask();
                task.Title = string.Format(CultureInfo.CurrentCulture, "Upload Now ({0})", account.Name);
                task.ApplicationPath = App.ExecutablePath;
                task.Arguments = Program.GetCommandLineArgumentsForUploadUI(account.Name);
                task.Description = string.Format(CultureInfo.CurrentCulture, "Upload the next batch of pictures for the \"{0}\" account now", account.Name);
                jumpList.JumpItems.Add(task);
            }
            jumpList.JumpItems.Add(new JumpTask { Title = "Open Log File", ApplicationPath = "notepad.exe", Arguments = PathProvider.LogFilePath, Description = "Open the log file" });
            jumpList.JumpItems.Add(new JumpTask { Title = App.ApplicationName + " Homepage", ApplicationPath = App.ApplicationUrl, Description = "Go to the homepage for " + App.ApplicationName });
            jumpList.Apply();
            JumpList.SetJumpList(this, jumpList);
        }

        #endregion

        #region Commands

        private bool CanOpenLogFile(object parameter)
        {
            return (File.Exists(PathProvider.LogFilePath));
        }

        private void OpenLogFile(object parameter)
        {
            Process.Start(PathProvider.LogFilePath);
        }

        private bool CanSaveChanges(object parameter)
        {
            return true;
        }

        private void SaveChanges(object parameter)
        {
            Messenger.Send<StatusMessage>(new StatusMessage("All changes are saved"));
            SchedulrConfigurationProvider.Save(this.Configuration);
        }

        private bool CanUndoChanges(object parameter)
        {
            return true;
        }

        private void UndoChanges(object parameter)
        {
            var result = MessageBox.Show(App.Current.MainWindow, "Are you sure you want to undo all your changes?", "Undo changes?", MessageBoxButton.YesNo);
            if (result == MessageBoxResult.Yes)
            {
                Messenger.Send<StatusMessage>(new StatusMessage("All changes are undone"));
                this.Configuration = SchedulrConfigurationProvider.LoadOrCreate();
            }
        }

        private bool CanExportConfiguration(object parameter)
        {
            return true;
        }

        private void ExportConfiguration(object parameter)
        {
            var saveFileDialog = new SaveFileDialog();
            saveFileDialog.Title = "Please select the configuration file to export to";
            saveFileDialog.Filter = "XML files|*.xml|All files|*.*";
            if (saveFileDialog.ShowDialog() == true)
            {
                try
                {
                    var fileName = saveFileDialog.FileName;
                    Messenger.Send<StatusMessage>(new StatusMessage(string.Format(CultureInfo.CurrentCulture, "Configuration exported to \"{0}\"", Path.GetFileName(fileName))));
                    SchedulrConfigurationProvider.Save(this.Configuration, fileName);
                }
                catch (Exception)
                {
                    MessageBox.Show(App.Current.MainWindow, string.Format(CultureInfo.CurrentCulture, "The specified configuration file could not be exported.{0}{0}Please see the log file for more details.", Environment.NewLine), "Error exporting configuration", MessageBoxButton.OK, MessageBoxImage.Warning);
                }
            }
        }

        private bool CanImportConfiguration(object parameter)
        {
            return true;
        }

        private void ImportConfiguration(object parameter)
        {
            var openFileDialog = new OpenFileDialog();
            openFileDialog.Title = "Please select the configuration file to import";
            openFileDialog.Filter = "XML files|*.xml|All files|*.*";
            if (openFileDialog.ShowDialog() == true)
            {
                try
                {
                    var fileName = openFileDialog.FileName;
                    Messenger.Send<StatusMessage>(new StatusMessage(string.Format(CultureInfo.CurrentCulture, "Configuration imported from \"{0}\"", Path.GetFileName(fileName))));
                    this.Configuration = SchedulrConfigurationProvider.Load(fileName);
                }
                catch (Exception)
                {
                    MessageBox.Show(App.Current.MainWindow, string.Format(CultureInfo.CurrentCulture, "The specified configuration file could not be imported.{0}{0}Please see the log file for more details.", Environment.NewLine), "Error importing configuration", MessageBoxButton.OK, MessageBoxImage.Warning);
                }
            }
        }

        #endregion

        #region Helper Methods

        /// <summary>
        /// Gets the OPC pack URL to a relative application resource.
        /// </summary>
        /// <param name="relativeResourcePath">The relative path to the embedded resource.</param>
        /// <returns>The OPC pack URL to the resource.</returns>
        internal static string GetResourceUrl(string relativeResourcePath)
        {
            return "pack://application:,,," + relativeResourcePath;
        }

        internal static bool IsVideoFile(string fileName)
        {
            if (string.IsNullOrWhiteSpace(fileName))
            {
                return false;
            }

            // Flickr officially supports JPEGs, non-animated GIFs, and PNGs.
            // You can also upload TIFFs and some other file types, but they will automatically be converted to and stored in JPEG format.

            // Flickr accepts the following video file formats:
            //     AVI (Proprietary codecs may not work)
            //     WMV
            //     MOV (AVID or other proprietary codecs may not work)
            //     MPEG (1, 2, and 4)
            //     3gp
            //     M2TS
            //     OGG
            //     OGV

            // Check if the file extension is any of the known video file extensions.
            var extension = Path.GetExtension(fileName);
            return App.VideoFileExtensions.Any(e => string.Equals(extension, e, StringComparison.InvariantCultureIgnoreCase));
        }

        #endregion
    }
}