﻿using System.Collections.ObjectModel;
using System.Runtime.Serialization;
using JelleDruyts.Windows.ObjectModel;

namespace Schedulr.Models
{
    /// <summary>
    /// Contains all settings for an <see cref="Account"/>.
    /// </summary>
    [DataContract(Namespace = Constants.DataContractNamespace)]
    public class AccountSettings : ObservableObject
    {
        #region Observable Properties

        /// <summary>
        /// Gets or sets the default values for new pictures.
        /// </summary>
        [DataMember]
        public Picture PictureDefaults
        {
            get { return this.GetValue(PictureDefaultsProperty); }
            set { this.SetValue(PictureDefaultsProperty, value); }
        }

        /// <summary>
        /// The definition for the <see cref="PictureDefaults"/> observable property.
        /// </summary>
        public static ObservableProperty<Picture> PictureDefaultsProperty = new ObservableProperty<Picture, AccountSettings>(o => o.PictureDefaults);

        /// <summary>
        /// Gets or sets a value that determines if picture metadata should be retrieved for new pictures.
        /// </summary>
        [DataMember]
        public bool RetrieveMetadata
        {
            get { return this.GetValue(RetrieveMetadataProperty); }
            set { this.SetValue(RetrieveMetadataProperty, value); }
        }

        /// <summary>
        /// The definition for the <see cref="RetrieveMetadata"/> observable property.
        /// </summary>
        public static ObservableProperty<bool> RetrieveMetadataProperty = new ObservableProperty<bool, AccountSettings>(o => o.RetrieveMetadata, true);

        /// <summary>
        /// Gets or sets a value that determines if pictures should be deleted after they have been uploaded.
        /// </summary>
        [DataMember]
        public bool DeleteFileAfterUpload
        {
            get { return this.GetValue(DeleteFileAfterUploadProperty); }
            set { this.SetValue(DeleteFileAfterUploadProperty, value); }
        }

        /// <summary>
        /// The definition for the <see cref="DeleteFileAfterUpload"/> observable property.
        /// </summary>
        public static ObservableProperty<bool> DeleteFileAfterUploadProperty = new ObservableProperty<bool, AccountSettings>(o => o.DeleteFileAfterUpload);

        /// <summary>
        /// Gets or sets the display mode of picture collections.
        /// </summary>
        [DataMember]
        public PictureCollectionDisplayMode PictureCollectionDisplayMode
        {
            get { return this.GetValue(PictureCollectionDisplayModeProperty); }
            set { this.SetValue(PictureCollectionDisplayModeProperty, value); }
        }

        /// <summary>
        /// The definition for the <see cref="PictureCollectionDisplayMode"/> observable property.
        /// </summary>
        public static ObservableProperty<PictureCollectionDisplayMode> PictureCollectionDisplayModeProperty = new ObservableProperty<PictureCollectionDisplayMode, AccountSettings>(o => o.PictureCollectionDisplayMode, PictureCollectionDisplayMode.TextAndIcon);

        /// <summary>
        /// Gets or sets the folders to monitor for new pictures.
        /// </summary>
        [DataMember]
        public ObservableCollection<string> FoldersToMonitor
        {
            get { return this.GetValue(FoldersToMonitorProperty); }
            set { this.SetValue(FoldersToMonitorProperty, value); }
        }

        /// <summary>
        /// The definition for the <see cref="FoldersToMonitor"/> observable property.
        /// </summary>
        public static ObservableProperty<ObservableCollection<string>> FoldersToMonitorProperty = new ObservableProperty<ObservableCollection<string>, AccountSettings>(o => o.FoldersToMonitor);

        /// <summary>
        /// Gets or sets the display mode of the picture queue.
        /// </summary>
        [DataMember]
        public PictureQueueDisplayMode PictureQueueDisplayMode
        {
            get { return this.GetValue(PictureQueueDisplayModeProperty); }
            set { this.SetValue(PictureQueueDisplayModeProperty, value); }
        }

        /// <summary>
        /// The definition for the <see cref="PictureQueueDisplayMode"/> observable property.
        /// </summary>
        public static ObservableProperty<PictureQueueDisplayMode> PictureQueueDisplayModeProperty = new ObservableProperty<PictureQueueDisplayMode, AccountSettings>(o => o.PictureQueueDisplayMode, PictureQueueDisplayMode.Vertical);

        /// <summary>
        /// Gets or sets the display mode of the picture previews.
        /// </summary>
        [DataMember]
        public PicturePreviewDisplayMode PicturePreviewDisplayMode
        {
            get { return this.GetValue(PicturePreviewDisplayModeProperty); }
            set { this.SetValue(PicturePreviewDisplayModeProperty, value); }
        }

        /// <summary>
        /// The definition for the <see cref="PicturePreviewDisplayMode"/> observable property.
        /// </summary>
        public static ObservableProperty<PicturePreviewDisplayMode> PicturePreviewDisplayModeProperty = new ObservableProperty<PicturePreviewDisplayMode, AccountSettings>(o => o.PicturePreviewDisplayMode, PicturePreviewDisplayMode.Thumbnail);

        /// <summary>
        /// Gets or sets the last selected geographic location.
        /// </summary>
        [DataMember]
        public GeoLocation LastLocation
        {
            get { return this.GetValue(LastLocationProperty); }
            set { this.SetValue(LastLocationProperty, value); }
        }

        /// <summary>
        /// The definition for the <see cref="LastLocation"/> observable property.
        /// </summary>
        public static ObservableProperty<GeoLocation> LastLocationProperty = new ObservableProperty<GeoLocation, AccountSettings>(o => o.LastLocation);

        /// <summary>
        /// Gets or sets the UI settings for the queued picture details.
        /// </summary>
        [DataMember]
        public PictureDetailsUISettings QueuedPictureDetailsUISettings
        {
            get { return this.GetValue(QueuedPictureDetailsUISettingsProperty); }
            set { this.SetValue(QueuedPictureDetailsUISettingsProperty, value); }
        }

        /// <summary>
        /// The definition for the <see cref="QueuedPictureDetailsUISettings"/> observable property.
        /// </summary>
        public static ObservableProperty<PictureDetailsUISettings> QueuedPictureDetailsUISettingsProperty = new ObservableProperty<PictureDetailsUISettings, AccountSettings>(o => o.QueuedPictureDetailsUISettings);

        /// <summary>
        /// Gets or sets the UI settings for the uploaded picture details.
        /// </summary>
        [DataMember]
        public PictureDetailsUISettings UploadedPictureDetailsUISettings
        {
            get { return this.GetValue(UploadedPictureDetailsUISettingsProperty); }
            set { this.SetValue(UploadedPictureDetailsUISettingsProperty, value); }
        }

        /// <summary>
        /// The definition for the <see cref="UploadedPictureDetailsUISettings"/> observable property.
        /// </summary>
        public static ObservableProperty<PictureDetailsUISettings> UploadedPictureDetailsUISettingsProperty = new ObservableProperty<PictureDetailsUISettings, AccountSettings>(o => o.UploadedPictureDetailsUISettings);

        /// <summary>
        /// Gets or sets the UI settings for the picture defaults details.
        /// </summary>
        [DataMember]
        public PictureDetailsUISettings PictureDefaultsDetailsUISettings
        {
            get { return this.GetValue(PictureDefaultsDetailsUISettingsProperty); }
            set { this.SetValue(PictureDefaultsDetailsUISettingsProperty, value); }
        }

        /// <summary>
        /// The definition for the <see cref="PictureDefaultsDetailsUISettings"/> observable property.
        /// </summary>
        public static ObservableProperty<PictureDetailsUISettings> PictureDefaultsDetailsUISettingsProperty = new ObservableProperty<PictureDetailsUISettings, AccountSettings>(o => o.PictureDefaultsDetailsUISettings);

        #endregion

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="AccountSettings"/> class.
        /// </summary>
        public AccountSettings()
        {
            this.PictureDefaults = new Picture();
            this.FoldersToMonitor = new ObservableCollection<string>();
            this.QueuedPictureDetailsUISettings = new PictureDetailsUISettings();
            this.UploadedPictureDetailsUISettings = new PictureDetailsUISettings();
            this.PictureDefaultsDetailsUISettings = new PictureDetailsUISettings();
        }

        #endregion
    }
}