﻿using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Windows;
using System.Windows.Data;
using System.Windows.Input;
using JelleDruyts.Windows;
using JelleDruyts.Windows.Collections;
using JelleDruyts.Windows.ObjectModel;
using Schedulr.Infrastructure;
using Schedulr.Messages;
using Schedulr.Models;

namespace Schedulr.ViewModels
{
    public abstract class PictureListViewModel : ViewModel
    {
        #region Fields

        private ObservableCollectionSynchronizer<Picture, PictureViewModel> synchronizer;
        private BatchIdGroupDescription batchIdGroupDescription;

        #endregion

        #region Properties

        protected RelayCommand RemoveCommand { get; private set; }
        protected RelayCommand ExpandAllBatchesCommand { get; private set; }
        protected RelayCommand CollapseAllBatchesCommand { get; private set; }
        protected Account Account { get; private set; }
        protected ObservableCollection<Picture> PictureList { get; private set; }
        protected ObservableCollection<PictureViewModel> PictureListViewModels { get; private set; }
        protected IList<Picture> SelectedPictures { get; private set; }
        public PictureQueueViewModel PictureQueue { get; private set; }

        #endregion

        #region Observable Properties

        /// <summary>
        /// Gets or sets the picture details view model.
        /// </summary>
        public PictureDetailsViewModel PictureDetails
        {
            get { return this.GetValue(PictureDetailsProperty); }
            set { this.SetValue(PictureDetailsProperty, value); }
        }

        /// <summary>
        /// The definition for the <see cref="PictureDetails"/> observable property.
        /// </summary>
        public static ObservableProperty<PictureDetailsViewModel> PictureDetailsProperty = new ObservableProperty<PictureDetailsViewModel, PictureListViewModel>(o => o.PictureDetails);

        #endregion

        #region Constructors

        protected PictureListViewModel()
        {
            this.RemoveCommand = new RelayCommand(this.Remove, this.CanRemove, "_Remove", "Removes the currently selected files from the queue [DEL]", new KeyGesture(Key.Delete));
            this.ExpandAllBatchesCommand = new RelayCommand(this.ExpandAllBatches, this.CanExpandAllBatches, "E_xpand All", "Expands all batches in the queue");
            this.CollapseAllBatchesCommand = new RelayCommand(this.CollapseAllBatches, this.CanCollapseAllBatches, "Co_llapse All", "Collapses all batches in the queue");
            this.PictureQueue = new PictureQueueViewModel(this.IsDragDropSupportedOnQueue, this.GetQueueCommands());
            this.SelectedPictures = new Picture[0];
            this.PictureList = new ObservableCollection<Picture>();
            Messenger.Register<AccountActionMessage>(OnAccountActionMessage);
            Messenger.Register<PictureQueueSelectionChangedMessage>(OnPictureQueueSelectionChanged);
            Messenger.Register<PictureQueueChangedMessage>(OnPictureQueueChanged);
        }

        #endregion

        #region Message Handlers

        private void OnAccountActionMessage(AccountActionMessage message)
        {
            if (message.Action == ListAction.CurrentChanged)
            {
                Logger.Log("PictureListViewModel - Current account changed", TraceEventType.Verbose);
                this.Account = message.Account;
                if (this.Account != null)
                {
                    this.PictureList = GetPicturesForAccount(this.Account);
                }
                else
                {
                    this.PictureList = new ObservableCollection<Picture>();
                }
                this.PictureQueue.Account = this.Account;
                Logger.Log("PictureListViewModel - PicturesList created", TraceEventType.Verbose);

                var groupedPicturesView = new CollectionViewSource();
                if (this.PictureListViewModels != null)
                {
                    foreach (var pictureListViewModel in this.PictureListViewModels)
                    {
                        pictureListViewModel.Dispose();
                    }
                    Logger.Log("PictureListViewModel - Disposed previous PictureListViewModels", TraceEventType.Verbose);
                }
                var pictureListViewModels = new ObservableCollection<PictureViewModel>(this.PictureList.Select(p => new PictureViewModel(p, this.Account)));
                Logger.Log("PictureListViewModel - Created current PictureListViewModels", TraceEventType.Verbose);
                this.PictureListViewModels = pictureListViewModels;
                Logger.Log("PictureListViewModel - Set current PictureListViewModels", TraceEventType.Verbose);
                groupedPicturesView.Source = this.PictureListViewModels;
                Logger.Log("PictureListViewModel - Set Source for GroupedPicturesView", TraceEventType.Verbose);
                this.batchIdGroupDescription = null;
                if (this.Account != null)
                {
                    this.batchIdGroupDescription = new BatchIdGroupDescription(this.Account, this.PictureList);
                    groupedPicturesView.GroupDescriptions.Add(this.batchIdGroupDescription);
                    Logger.Log("PictureListViewModel - Set GroupDescription for GroupedPicturesView", TraceEventType.Verbose);
                }
                this.PictureQueue.Pictures = groupedPicturesView.View;
                Logger.Log("PictureListViewModel - Set final GroupedPicturesView to queue", TraceEventType.Verbose);

                // Dispose the previous synchronizer if there was one (to avoid leaking event handlers).
                if (this.synchronizer != null)
                {
                    this.synchronizer.Dispose();
                }
                // Synchronize the picture view models with the source pictures.
                this.synchronizer = new ObservableCollectionSynchronizer<Picture, PictureViewModel>(this.PictureList, p => new PictureViewModel(p, this.Account), this.PictureListViewModels, p => p.Picture);
                Logger.Log("PictureListViewModel - Set up synchronizer", TraceEventType.Verbose);

                SetPictureDetails();
            }
        }

        private void SetPictureDetails()
        {
            this.PictureDetails = new PictureDetailsViewModel(this.Account, this.SelectedPictures, GetUISettingsForAccount(this.Account), Visibility.Visible, Visibility.Visible, this.ArePictureDetailsReadOnly);
        }

        private void OnPictureQueueSelectionChanged(PictureQueueSelectionChangedMessage message)
        {
            if (message.PictureQueue == this.PictureQueue)
            {
                this.SelectedPictures = message.SelectedPictures;
                SetPictureDetails();
            }
        }

        private void OnPictureQueueChanged(PictureQueueChangedMessage message)
        {
            if (message.Reason == PictureQueueChangedReason.BatchesNormalized || message.Reason == PictureQueueChangedReason.PicturesShuffled)
            {
                if (this.PictureQueue != null && this.PictureQueue.Pictures != null)
                {
                    this.PictureQueue.Pictures.Refresh();
                }
            }
        }

        #endregion

        #region Commands

        private bool CanRemove(object parameter)
        {
            return (this.SelectedPictures.Count > 0);
        }

        private void Remove(object parameter)
        {
            var removingPicturesTask = new ApplicationTask("Removing files", this.SelectedPictures.Count);
            var step = 0;
            Messenger.Send<TaskStatusMessage>(new TaskStatusMessage(removingPicturesTask));
            try
            {
                var i = 0;
                foreach (var picture in this.SelectedPictures)
                {
                    removingPicturesTask.SetProgress(step, string.Format(CultureInfo.CurrentCulture, "Removing \"{0}\" ({1} of {2})", picture.Title, ++i, removingPicturesTask.TotalSteps));
                    step++;
                    this.PictureList.Remove(picture);
                }
                Tasks.NormalizeBatches(this.PictureList);
            }
            finally
            {
                removingPicturesTask.SetComplete(string.Format(CultureInfo.CurrentCulture, "{0} removed", removingPicturesTask.TotalSteps.Value.ToCountString("file")));
            }
        }

        private bool CanExpandAllBatches(object parameter)
        {
            return this.batchIdGroupDescription != null;
        }

        private void ExpandAllBatches(object parameter)
        {
            this.batchIdGroupDescription.SetBatchesExpanded(true);
        }

        private bool CanCollapseAllBatches(object parameter)
        {
            return this.batchIdGroupDescription != null;
        }

        private void CollapseAllBatches(object parameter)
        {
            this.batchIdGroupDescription.SetBatchesExpanded(false);
        }

        #endregion

        #region Abstract Members

        protected abstract ObservableCollection<Picture> GetPicturesForAccount(Account account);

        protected abstract PictureDetailsUISettings GetUISettingsForAccount(Account account);

        protected abstract bool IsDragDropSupportedOnQueue { get; }

        protected abstract IEnumerable<ICommand> GetQueueCommands();

        protected abstract bool ArePictureDetailsReadOnly { get; }

        #endregion
    }
}