﻿using System;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Text;
using System.Xml;
using JelleDruyts.Windows;
using Schedulr.Models;

namespace Schedulr.Providers
{
    /// <summary>
    /// Provides metadata services for pictures.
    /// </summary>
    public static class PictureMetadataProvider
    {
        #region RetrieveMetadata

        /// <summary>
        /// Retrieves the meta data.
        /// </summary>
        /// <param name="fileName">The file in which to find the embedded meta data XML.</param>
        /// <param name="picture">The picture to update with the retrieved metadata.</param>
        /// <param name="logger">The logger to use.</param>
        public static void RetrieveMetadataFromFile(string fileName, Picture picture, ILogger logger)
        {
            if (!File.Exists(fileName))
            {
                throw new ArgumentException("The specified file does not exist: " + fileName);
            }
            try
            {
                logger.Log(string.Format(CultureInfo.CurrentCulture, "Retrieving metadata from file \"{0}\"", fileName), TraceEventType.Information);
                string fileContents = null;
                using (var reader = new StreamReader(fileName))
                {
                    fileContents = reader.ReadToEnd();
                }
                const string beginCapture = "<rdf:RDF";
                const string endCapture = "</rdf:RDF>";
                var beginPos = fileContents.IndexOf(beginCapture, StringComparison.OrdinalIgnoreCase);
                var endPos = fileContents.IndexOf(endCapture, StringComparison.OrdinalIgnoreCase);
                if (beginPos >= 0 && endPos > beginPos)
                {
                    var xmlPart = fileContents.Substring(beginPos, (endPos - beginPos) + endCapture.Length);
                    XmlDocument doc = new XmlDocument();
                    doc.LoadXml(xmlPart);
                    RetrieveMetadataFromXml(doc, picture);
                }
                else
                {
                    logger.Log("Could not retrieve metadata from file " + fileName, TraceEventType.Information);
                }
            }
            catch (Exception exc)
            {
                logger.Log("Could not retrieve metadata from file " + fileName, exc);
            }
        }

        /// <summary>
        /// Retrieves the meta data.
        /// </summary>
        /// <param name="embeddedMetadataXml">The embedded meta data XML.</param>
        /// <param name="picture">The picture to update with the retrieved metadata.</param>
        private static void RetrieveMetadataFromXml(XmlDocument embeddedMetadataXml, Picture picture)
        {
            // Thank you Omar: http://www.shahine.com/omar/ReadingXMPMetadataFromAJPEGUsingC.aspx.
            XmlNamespaceManager nsMgr = new XmlNamespaceManager(embeddedMetadataXml.NameTable);
            nsMgr.AddNamespace("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
            nsMgr.AddNamespace("exif", "http://ns.adobe.com/exif/1.0/");
            nsMgr.AddNamespace("x", "adobe:ns:meta/");
            nsMgr.AddNamespace("xap", "http://ns.adobe.com/xap/1.0/");
            nsMgr.AddNamespace("tiff", "http://ns.adobe.com/tiff/1.0/");
            nsMgr.AddNamespace("dc", "http://purl.org/dc/elements/1.1/");

            // Get tags.
            XmlNode xmlNode = embeddedMetadataXml.SelectSingleNode("/rdf:RDF/rdf:Description/dc:subject/rdf:Bag", nsMgr);
            if (xmlNode != null)
            {
                StringBuilder tags = new StringBuilder();
                foreach (XmlNode tagNode in xmlNode)
                {
                    if (tags.Length > 0)
                    {
                        // Separate tags with spaces.
                        tags.Append(" ");
                    }
                    string tag = tagNode.InnerText;
                    if (tag.Contains(" "))
                    {
                        // If the tag contains spaces, put double quotes around them.
                        tag = string.Format(CultureInfo.CurrentCulture, "\"{0}\"", tag);
                    }
                    tags.Append(tag);
                }
                picture.Tags = tags.ToString();
            }

            // Get the title.
            xmlNode = embeddedMetadataXml.SelectSingleNode("/rdf:RDF/rdf:Description/dc:title/rdf:Alt", nsMgr);
            if (xmlNode != null && xmlNode.ChildNodes.Count > 0)
            {
                picture.Title = xmlNode.ChildNodes[0].InnerText;
            }

            // Get the description.
            xmlNode = embeddedMetadataXml.SelectSingleNode("/rdf:RDF/rdf:Description/dc:description/rdf:Alt", nsMgr);
            if (xmlNode != null && xmlNode.ChildNodes.Count > 0)
            {
                picture.Description = xmlNode.ChildNodes[0].InnerText;
            }

            // Get the GPS location.
            xmlNode = embeddedMetadataXml.SelectSingleNode("/rdf:RDF/rdf:Description/exif:GPSLatitude", nsMgr);
            if (xmlNode != null && !string.IsNullOrEmpty(xmlNode.InnerText))
            {
                double latitude;
                if (Sexagesimal.TryParse(xmlNode.InnerText, out latitude))
                {
                    xmlNode = embeddedMetadataXml.SelectSingleNode("/rdf:RDF/rdf:Description/exif:GPSLongitude", nsMgr);
                    if (xmlNode != null && !string.IsNullOrEmpty(xmlNode.InnerText))
                    {
                        double longitude;
                        if (Sexagesimal.TryParse(xmlNode.InnerText, out longitude))
                        {
                            picture.Location = new GeoLocation(latitude, longitude, GeoLocation.MaxAccuracy);
                        }
                    }
                }
            }
        }

        #endregion
    }
}